@@ -2,7 +2,7 @@ package PAGI::Server;
 use strict;
 use warnings;
 
-our $VERSION = '0.001002';
+our $VERSION = '0.001003';
 
 use parent 'IO::Async::Notifier';
 use IO::Async::Listener;
@@ -492,6 +492,34 @@ are faster than the overhead of async I/O.
 
 B<CLI:> C<--sync-file-threshold NUM>
 
+=item sendfile_timeout => $seconds
+
+Timeout in seconds for sendfile() socket writability. When using sendfile()
+for large file transfers, the socket buffer may fill up, requiring the server
+to wait for the client to drain it. This timeout prevents indefinite waits.
+
+B<Default:> 30 seconds
+
+This timeout protects against:
+
+=over 4
+
+=item * Slow or unresponsive clients that accept connections but read slowly
+
+=item * Network congestion causing socket buffer backpressure
+
+=item * Malicious clients attempting to tie up server resources
+
+=back
+
+B<Example:>
+
+    # Increase timeout for slow network conditions
+    my $server = PAGI::Server->new(
+        app => $app,
+        sendfile_timeout => 60,  # 60 seconds
+    );
+
 =item max_requests => $count
 
 Maximum number of requests a worker process will handle before restarting.
@@ -520,6 +548,117 @@ B<CLI:> C<--max-requests 10000>
 Example: With 4 workers and max_requests=10000, total capacity before any
 restart is 40,000 requests. Workers restart individually without downtime.
 
+=item request_timeout => $seconds
+
+Maximum time in seconds a request can stall without any I/O activity before
+being terminated. This is a "stall timeout" - the timer resets whenever data
+is read from the client or written to the client.
+
+B<Default:> 0 (disabled)
+
+B<Why disabled by default:> Creating per-request timers adds overhead that
+impacts throughput on high-performance workloads. For maximum performance,
+this is disabled by default. Most production deployments run behind a reverse
+proxy (nginx, haproxy) which provides its own timeout protection.
+
+B<When to enable:>
+
+=over 4
+
+=item * Running PAGI directly without a reverse proxy
+
+=item * Small/internal apps where simplicity matters more than max throughput
+
+=item * Untrusted clients that might send data slowly or hang
+
+=item * Defense against application bugs that cause requests to hang indefinitely
+
+=back
+
+B<How it works:>
+
+=over 4
+
+=item * Timer starts when request processing begins
+
+=item * Timer resets on any read activity (receiving request body)
+
+=item * Timer resets on any write activity (sending response)
+
+=item * If timer expires (no I/O for N seconds), connection is closed
+
+=item * Not used for WebSocket/SSE (they have C<ws_idle_timeout>/C<sse_idle_timeout>)
+
+=back
+
+B<Example:>
+
+    # Enable 30 second stall timeout (recommended when not behind proxy)
+    my $server = PAGI::Server->new(
+        app             => $app,
+        request_timeout => 30,
+    );
+
+B<CLI:> C<--request-timeout 30>
+
+B<Note:> This differs from C<timeout> (idle connection timeout). The
+C<timeout> applies between requests on keep-alive connections. The
+C<request_timeout> applies during active request processing.
+
+=item ws_idle_timeout => $seconds
+
+Maximum time in seconds a WebSocket connection can be idle without any
+activity (no messages sent or received) before being closed.
+
+B<Default:> 0 (disabled - WebSocket connections can be idle indefinitely)
+
+When enabled, the timer resets on:
+
+=over 4
+
+=item * Sending any WebSocket frame (accept, send, ping, close)
+
+=item * Receiving any WebSocket frame from client
+
+=back
+
+B<Example:>
+
+    # Close idle WebSocket connections after 5 minutes
+    my $server = PAGI::Server->new(
+        app             => $app,
+        ws_idle_timeout => 300,
+    );
+
+B<CLI:> C<--ws-idle-timeout 300>
+
+B<Note:> For more sophisticated keep-alive behavior with ping/pong, use
+the C<PAGI::Middleware::WebSocket::Heartbeat> middleware instead.
+
+=item sse_idle_timeout => $seconds
+
+Maximum time in seconds an SSE connection can be idle without any events
+being sent before being closed.
+
+B<Default:> 0 (disabled - SSE connections can be idle indefinitely)
+
+The timer resets each time an event is sent to the client (including
+comments and the initial headers).
+
+B<Example:>
+
+    # Close idle SSE connections after 2 minutes
+    my $server = PAGI::Server->new(
+        app              => $app,
+        sse_idle_timeout => 120,
+    );
+
+B<CLI:> C<--sse-idle-timeout 120>
+
+B<Note:> For SSE connections that may be legitimately idle, consider
+using the C<PAGI::Middleware::SSE::Heartbeat> middleware to send
+periodic comment keepalives.
+
 =back
 
 =head1 METHODS
@@ -656,6 +795,30 @@ the worker pool method, which is more portable but slightly slower:
 
 =back
 
+=head2 Troubleshooting Sendfile Issues
+
+B<Error: "sendfile write timeout after N seconds">
+
+This error occurs when the client cannot drain the socket buffer fast enough.
+Common causes and solutions:
+
+=over 4
+
+=item * B<Slow Clients:> Clients on slow networks may need more time. Increase
+the C<sendfile_timeout> parameter or use C<< disable_sendfile => 1 >> to use
+the more portable worker pool fallback.
+
+=item * B<Large Files:> Very large files may need longer timeouts. Consider
+using a dedicated file server (nginx, CDN) for large static assets.
+
+=back
+
+B<FreeBSD-Specific Notes:>
+
+FreeBSD's sendfile() implementation returns C<EAGAIN> with zero bytes sent
+when the socket buffer is completely full. PAGI handles this correctly, but
+constrained environments may hit the timeout more easily.
+
 =head1 ENABLING TLS SUPPORT
 
 PAGI::Server supports HTTPS/TLS connections, but requires additional modules
@@ -856,6 +1019,10 @@ sub _init {
     $self->{max_connections}     = delete $params->{max_connections} // 0;  # 0 = auto-detect
     $self->{disable_sendfile}    = delete $params->{disable_sendfile} // 0;  # Disable sendfile() syscall for file responses
     $self->{sync_file_threshold} = delete $params->{sync_file_threshold} // 65536;  # Threshold for sync file reads (0=always async)
+    $self->{sendfile_timeout}    = delete $params->{sendfile_timeout};  # Timeout for sendfile socket writability (undef = use Connection default)
+    $self->{request_timeout}     = delete $params->{request_timeout} // 0;  # Request stall timeout in seconds (0 = disabled, default for performance)
+    $self->{ws_idle_timeout}     = delete $params->{ws_idle_timeout} // 0;   # WebSocket idle timeout (0 = disabled)
+    $self->{sse_idle_timeout}    = delete $params->{sse_idle_timeout} // 0;  # SSE idle timeout (0 = disabled)
 
     $self->{running}     = 0;
     $self->{bound_port}  = undef;
@@ -944,6 +1111,18 @@ sub configure {
     if (exists $params{disable_sendfile}) {
         $self->{disable_sendfile} = delete $params{disable_sendfile};
     }
+    if (exists $params{sendfile_timeout}) {
+        $self->{sendfile_timeout} = delete $params{sendfile_timeout};
+    }
+    if (exists $params{request_timeout}) {
+        $self->{request_timeout} = delete $params{request_timeout};
+    }
+    if (exists $params{ws_idle_timeout}) {
+        $self->{ws_idle_timeout} = delete $params{ws_idle_timeout};
+    }
+    if (exists $params{sse_idle_timeout}) {
+        $self->{sse_idle_timeout} = delete $params{sse_idle_timeout};
+    }
 
     $self->SUPER::configure(%params);
 }
@@ -1123,9 +1302,15 @@ async sub _listen_singleworker {
         my $shutdown_handler = sub {
             return if $shutdown_triggered;
             $shutdown_triggered = 1;
-            $self->shutdown->on_done(sub {
-                $self->loop->stop;
-            })->retain;
+            $self->adopt_future(
+                $self->shutdown->on_done(sub {
+                    $self->loop->stop;
+                })->on_fail(sub {
+                    my ($error) = @_;
+                    $self->_log(error => "Shutdown error: $error");
+                    $self->loop->stop;  # Still stop even on error
+                })
+            );
         };
         $self->loop->watch_signal(TERM => $shutdown_handler);
         $self->loop->watch_signal(INT => $shutdown_handler);
@@ -1403,9 +1588,15 @@ sub _run_as_worker {
         $loop->watch_signal(TERM => sub {
             return if $shutdown_triggered;
             $shutdown_triggered = 1;
-            $worker_server->shutdown->on_done(sub {
-                $loop->stop;
-            })->retain;
+            $worker_server->adopt_future(
+                $worker_server->shutdown->on_done(sub {
+                    $loop->stop;
+                })->on_fail(sub {
+                    my ($error) = @_;
+                    $worker_server->_log(error => "Worker shutdown error: $error");
+                    $loop->stop;  # Still stop even on error
+                })
+            );
         });
     }
 
@@ -1413,7 +1604,7 @@ sub _run_as_worker {
     my $startup_done = 0;
     my $startup_error;
 
-    (async sub {
+    my $startup_future = (async sub {
         eval {
             my $startup_result = await $worker_server->_run_lifespan_startup;
             if (!$startup_result->{success}) {
@@ -1425,7 +1616,10 @@ sub _run_as_worker {
         }
         $startup_done = 1;
         $loop->stop if $startup_error;  # Stop loop on error
-    })->()->retain;
+    })->();
+
+    # Use adopt_future instead of retain
+    $worker_server->adopt_future($startup_future);
 
     # Run the loop briefly to let async startup complete
     $loop->loop_once while !$startup_done;
@@ -1495,12 +1689,16 @@ sub _on_connection {
         state             => $self->{state},
         tls_enabled       => $self->{tls_enabled} // 0,
         timeout           => $self->{timeout},
+        request_timeout   => $self->{request_timeout},
+        ws_idle_timeout   => $self->{ws_idle_timeout},
+        sse_idle_timeout  => $self->{sse_idle_timeout},
         max_body_size     => $self->{max_body_size},
         access_log        => $self->{access_log},
         max_receive_queue => $self->{max_receive_queue},
         max_ws_frame_size => $self->{max_ws_frame_size},
         disable_sendfile  => $self->{disable_sendfile},
         sync_file_threshold => $self->{sync_file_threshold},
+        (defined $self->{sendfile_timeout} ? (sendfile_timeout => $self->{sendfile_timeout}) : ()),
     );
 
     # Track the connection (O(1) hash insert)
@@ -1618,9 +1816,15 @@ sub _on_request_complete {
         $self->{_max_requests_shutdown_triggered} = 1;
         $self->_log(info => "Worker $$: reached max_requests ($self->{max_requests}), shutting down");
         # Initiate graceful shutdown (finish current connections, then exit)
-        $self->shutdown->on_done(sub {
-            $self->loop->stop;
-        })->retain;
+        $self->adopt_future(
+            $self->shutdown->on_done(sub {
+                $self->loop->stop;
+            })->on_fail(sub {
+                my ($error) = @_;
+                $self->_log(error => "Worker $$: max_requests shutdown error: $error");
+                $self->loop->stop;  # Still stop even on error
+            })
+        );
     }
 }
 
@@ -1719,7 +1923,8 @@ async sub _run_lifespan_startup {
 
     # Keep the app future so we can trigger shutdown later
     $self->{lifespan_app_future} = $app_future;
-    $app_future->retain;
+    # Use adopt_future instead of retain for proper error handling
+    $self->adopt_future($app_future);
 
     # Wait for startup complete (with timeout)
     my $result = await $startup_complete;
