@@ -14,6 +14,7 @@ use PAGI::Util::AsyncFile;
 
 
 use constant FILE_CHUNK_SIZE => 65536;  # 64KB chunks for file streaming
+use constant DEFAULT_SENDFILE_TIMEOUT => 30;  # Default timeout for sendfile socket writability (seconds)
 
 # Try to load Sys::Sendfile for zero-copy file transfers
 # Falls back to chunked reads if not available
@@ -105,6 +106,17 @@ connection. It handles:
 
 =back
 
+=head2 Sendfile Timeout Configuration
+
+When using sendfile() for large file transfers, the socket buffer may fill
+up faster than the client can drain it. The C<sendfile_timeout> parameter
+controls how long to wait for the socket to become writable before giving up.
+
+B<Default:> 30 seconds
+
+Configure via the C<sendfile_timeout> parameter in L<PAGI::Server> or
+directly in the Connection constructor.
+
 =cut
 
 sub new {
@@ -119,12 +131,16 @@ sub new {
         state         => $args{state} // {},
         tls_enabled   => $args{tls_enabled} // 0,
         timeout       => $args{timeout} // 60,  # Idle timeout in seconds
+        request_timeout => $args{request_timeout} // 0,  # Request stall timeout in seconds (0 = disabled, default for performance)
+        ws_idle_timeout => $args{ws_idle_timeout} // 0,   # WebSocket idle timeout (0 = disabled)
+        sse_idle_timeout => $args{sse_idle_timeout} // 0,  # SSE idle timeout (0 = disabled)
         max_body_size     => $args{max_body_size},  # 0 = unlimited
         access_log        => $args{access_log},     # Filehandle for access logging
         max_receive_queue => $args{max_receive_queue} // 1000,  # Max WebSocket receive queue size
         max_ws_frame_size => $args{max_ws_frame_size} // 65536,  # Max WebSocket frame size in bytes
         disable_sendfile  => $args{disable_sendfile} // 0,  # Disable sendfile even if available
         sync_file_threshold => $args{sync_file_threshold} // 65536,  # Threshold for sync file reads (0=always async)
+        sendfile_timeout  => $args{sendfile_timeout} // DEFAULT_SENDFILE_TIMEOUT,  # Timeout for sendfile socket writability (seconds)
         tls_info      => undef,  # Populated on first request if TLS
         buffer        => '',
         closed        => 0,
@@ -132,6 +148,9 @@ sub new {
         response_status  => undef,  # Track response status for logging
         request_start    => undef,  # Track request start time for logging
         idle_timer    => undef,  # IO::Async::Timer for idle timeout
+        stall_timer   => undef,  # IO::Async::Timer for request stall timeout
+        ws_idle_timer => undef,  # IO::Async::Timer for WebSocket idle timeout
+        sse_idle_timer => undef, # IO::Async::Timer for SSE idle timeout
         # Event queue for $receive
         receive_queue   => [],
         receive_pending => undef,
@@ -214,6 +233,9 @@ sub start {
             # Reset idle timer on any read activity
             $weak_self->_reset_idle_timer;
 
+            # Reset stall timer on read activity (if handling a request)
+            $weak_self->_reset_stall_timer if $weak_self->{handling_request};
+
             $weak_self->{buffer} .= $$buffref;
             $$buffref = '';
 
@@ -276,6 +298,145 @@ sub _stop_idle_timer {
     $self->{idle_timer} = undef;
 }
 
+# Request stall timeout - closes connection if no I/O activity during request processing
+sub _start_stall_timer {
+    my ($self) = @_;
+
+    return unless $self->{request_timeout} && $self->{request_timeout} > 0;
+    return unless $self->{server};
+    return if $self->{stall_timer};  # Already running
+
+    weaken(my $weak_self = $self);
+
+    my $timer = IO::Async::Timer::Countdown->new(
+        delay => $self->{request_timeout},
+        on_expire => sub {
+            return unless $weak_self;
+            return if $weak_self->{closed};
+            # Log the timeout
+            if ($weak_self->{server} && $weak_self->{server}->can('_log')) {
+                $weak_self->{server}->_log(warn =>
+                    "Request stall timeout ($weak_self->{request_timeout}s) - closing connection");
+            }
+            $weak_self->_close;
+        },
+    );
+    $self->{stall_timer} = $timer;
+    $self->{server}->add_child($timer);
+    $timer->start;
+}
+
+sub _reset_stall_timer {
+    my ($self) = @_;
+
+    return unless $self->{stall_timer};
+    $self->{stall_timer}->reset;
+    $self->{stall_timer}->start unless $self->{stall_timer}->is_running;
+}
+
+sub _stop_stall_timer {
+    my ($self) = @_;
+
+    return unless $self->{stall_timer};
+    $self->{stall_timer}->stop if $self->{stall_timer}->is_running;
+    if ($self->{server}) {
+        $self->{server}->remove_child($self->{stall_timer});
+    }
+    $self->{stall_timer} = undef;
+}
+
+# WebSocket idle timeout - closes connection if no activity
+sub _start_ws_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{ws_idle_timeout} && $self->{ws_idle_timeout} > 0;
+    return unless $self->{server};
+    return if $self->{ws_idle_timer};
+
+    weaken(my $weak_self = $self);
+
+    my $timer = IO::Async::Timer::Countdown->new(
+        delay => $self->{ws_idle_timeout},
+        on_expire => sub {
+            return unless $weak_self;
+            return if $weak_self->{closed};
+            if ($weak_self->{server} && $weak_self->{server}->can('_log')) {
+                $weak_self->{server}->_log(warn =>
+                    "WebSocket idle timeout ($weak_self->{ws_idle_timeout}s) - closing connection");
+            }
+            $weak_self->_close;
+        },
+    );
+    $self->{ws_idle_timer} = $timer;
+    $self->{server}->add_child($timer);
+    $timer->start;
+}
+
+sub _reset_ws_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{ws_idle_timer};
+    $self->{ws_idle_timer}->reset;
+    $self->{ws_idle_timer}->start unless $self->{ws_idle_timer}->is_running;
+}
+
+sub _stop_ws_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{ws_idle_timer};
+    $self->{ws_idle_timer}->stop if $self->{ws_idle_timer}->is_running;
+    if ($self->{server}) {
+        $self->{server}->remove_child($self->{ws_idle_timer});
+    }
+    $self->{ws_idle_timer} = undef;
+}
+
+# SSE idle timeout - closes connection if no activity
+sub _start_sse_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{sse_idle_timeout} && $self->{sse_idle_timeout} > 0;
+    return unless $self->{server};
+    return if $self->{sse_idle_timer};
+
+    weaken(my $weak_self = $self);
+
+    my $timer = IO::Async::Timer::Countdown->new(
+        delay => $self->{sse_idle_timeout},
+        on_expire => sub {
+            return unless $weak_self;
+            return if $weak_self->{closed};
+            if ($weak_self->{server} && $weak_self->{server}->can('_log')) {
+                $weak_self->{server}->_log(warn =>
+                    "SSE idle timeout ($weak_self->{sse_idle_timeout}s) - closing connection");
+            }
+            $weak_self->_close;
+        },
+    );
+    $self->{sse_idle_timer} = $timer;
+    $self->{server}->add_child($timer);
+    $timer->start;
+}
+
+sub _reset_sse_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{sse_idle_timer};
+    $self->{sse_idle_timer}->reset;
+    $self->{sse_idle_timer}->start unless $self->{sse_idle_timer}->is_running;
+}
+
+sub _stop_sse_idle_timer {
+    my ($self) = @_;
+
+    return unless $self->{sse_idle_timer};
+    $self->{sse_idle_timer}->stop if $self->{sse_idle_timer}->is_running;
+    if ($self->{server}) {
+        $self->{server}->remove_child($self->{sse_idle_timer});
+    }
+    $self->{sse_idle_timer} = undef;
+}
+
 sub _try_handle_request {
     my ($self) = @_;
 
@@ -322,12 +483,14 @@ sub _try_handle_request {
     } elsif ($is_sse) {
         $self->{request_future} = $self->_handle_sse_request($request);
     } else {
+        # Start stall timer for HTTP requests (WebSocket/SSE have their own handling)
+        $self->_start_stall_timer;
         $self->{request_future} = $self->_handle_request($request);
     }
 
-    # Retain the future to prevent warning if connection is destroyed
-    # while request is still being processed
-    $self->{request_future}->retain;
+    # Use adopt_future for proper error tracking instead of retain
+    # This ensures errors are propagated to the server's error handling
+    $self->{server}->adopt_future($self->{request_future});
 }
 
 sub _is_websocket_upgrade {
@@ -410,6 +573,9 @@ async sub _handle_request {
     # Notify server that request completed (for max_requests tracking)
     $self->{server}->_on_request_complete if $self->{server};
 
+    # Stop stall timer - request completed successfully
+    $self->_stop_stall_timer;
+
     # Determine if we should keep the connection alive
     my $keep_alive = $self->_should_keep_alive($request);
 
@@ -718,6 +884,9 @@ sub _create_send {
         return Future->done unless $weak_self;
         return Future->done if $weak_self->{closed};
 
+        # Reset stall timer on write activity
+        $weak_self->_reset_stall_timer;
+
         my $type = $event->{type} // '';
 
         if ($type eq 'http.response.start') {
@@ -997,6 +1166,13 @@ sub _close {
     # Stop idle timer
     $self->_stop_idle_timer;
 
+    # Stop stall timer
+    $self->_stop_stall_timer;
+
+    # Stop WS/SSE idle timers
+    $self->_stop_ws_idle_timer;
+    $self->_stop_sse_idle_timer;
+
     # Complete any pending receive with disconnect
     $self->_handle_disconnect;
 
@@ -1190,6 +1366,7 @@ async sub _handle_sse_request {
 
     $self->{sse_mode} = 1;
     $self->_stop_idle_timer;  # SSE connections are long-lived
+    $self->_start_sse_idle_timer;  # Start SSE-specific idle timer if configured
 
     my $scope = $self->_create_sse_scope($request);
     my $receive = $self->_create_sse_receive($request);
@@ -1208,7 +1385,9 @@ async sub _handle_sse_request {
     }
 
     # Send chunked terminator if SSE was started (uses chunked encoding)
-    if ($self->{sse_started} && !$self->{closed}) {
+    # Check both closed flag and that stream is still writable
+    if ($self->{sse_started} && !$self->{closed} &&
+        $self->{stream} && $self->{stream}->write_handle) {
         $self->{stream}->write("0\r\n\r\n");
     }
 
@@ -1308,6 +1487,9 @@ sub _create_sse_send {
         return Future->done unless $weak_self;
         return Future->done if $weak_self->{closed};
 
+        # Reset SSE idle timer on send activity
+        $weak_self->_reset_sse_idle_timer;
+
         my $type = $event->{type} // '';
 
         if ($type eq 'sse.start') {
@@ -1423,6 +1605,7 @@ async sub _handle_websocket_request {
     my ($self, $request) = @_;
 
     $self->_stop_idle_timer;  # WebSocket connections are long-lived
+    $self->_start_ws_idle_timer;  # Start WebSocket-specific idle timer if configured
 
     my $scope = $self->_create_websocket_scope($request);
     my $receive = $self->_create_websocket_receive($request);
@@ -1577,6 +1760,9 @@ sub _create_websocket_send {
         return Future->done unless $weak_self;
         return Future->done if $weak_self->{closed};
 
+        # Reset WebSocket idle timer on send activity
+        $weak_self->_reset_ws_idle_timer;
+
         my $type = $event->{type} // '';
 
         if ($type eq 'websocket.accept') {
@@ -1694,6 +1880,9 @@ sub _process_websocket_frames {
     return unless $self->{websocket_mode};
     return if $self->{closed};
 
+    # Reset WebSocket idle timer on receive activity
+    $self->_reset_ws_idle_timer;
+
     my $frame = $self->{websocket_frame};
 
     # Append buffer to frame parser
@@ -1863,7 +2052,7 @@ sub _can_use_sendfile {
 # Wait for socket to become writable (non-blocking, with timeout)
 # Used by sendfile loop when socket buffer is full (EAGAIN condition)
 # Returns a Future that resolves when socket is writable or fails on timeout/error
-use constant SENDFILE_WRITE_TIMEOUT => 30;  # seconds
+use constant DEFAULT_SENDFILE_TIMEOUT => 30;  # seconds
 
 sub _wait_for_writable {
     my ($self, $socket) = @_;
@@ -1890,8 +2079,9 @@ sub _wait_for_writable {
     $watching = 1;
 
     # Set up timeout to prevent infinite wait
-    my $timeout_f = $loop->delay_future(after => SENDFILE_WRITE_TIMEOUT)->then(sub {
-        Future->fail('sendfile write timeout after ' . SENDFILE_WRITE_TIMEOUT . ' seconds');
+    my $timeout = $self->{sendfile_timeout};
+    my $timeout_f = $loop->delay_future(after => $timeout)->then(sub {
+        Future->fail("sendfile write timeout after $timeout seconds");
     });
 
     # Race between write ready and timeout
