@@ -21,7 +21,12 @@ sub new {
     croak "PAGI::WebSocket requires scope type 'websocket', got '$scope->{type}'"
         unless ($scope->{type} // '') eq 'websocket';
 
-    return bless {
+    # Return existing WebSocket object if one was already created for this scope
+    # This ensures consistent state (is_connected, is_closed, callbacks) if
+    # multiple code paths create WebSocket objects from the same scope.
+    return $scope->{'pagi.websocket'} if $scope->{'pagi.websocket'};
+
+    my $self = bless {
         scope   => $scope,
         receive => $receive,
         send    => $send,
@@ -32,6 +37,12 @@ sub new {
         _on_error     => [],
         _on_message   => [],
     }, $class;
+
+    # Cache in scope for reuse (weakened to avoid circular reference leak)
+    $scope->{'pagi.websocket'} = $self;
+    Scalar::Util::weaken($scope->{'pagi.websocket'});
+
+    return $self;
 }
 
 # Scope property accessors
@@ -84,18 +95,18 @@ sub header {
     return $value;
 }
 
-# All headers as Hash::MultiValue (cached)
+# All headers as Hash::MultiValue (cached in scope)
 sub headers {
     my $self = shift;
-    return $self->{_headers} if $self->{_headers};
+    return $self->{scope}{'pagi.request.headers'} if $self->{scope}{'pagi.request.headers'};
 
     my @pairs;
     for my $pair (@{$self->{scope}{headers} // []}) {
         push @pairs, lc($pair->[0]), $pair->[1];
     }
 
-    $self->{_headers} = Hash::MultiValue->new(@pairs);
-    return $self->{_headers};
+    $self->{scope}{'pagi.request.headers'} = Hash::MultiValue->new(@pairs);
+    return $self->{scope}{'pagi.request.headers'};
 }
 
 # All values for a header
@@ -789,6 +800,12 @@ Creates a new WebSocket wrapper. Requires:
 
 Dies if scope type is not 'websocket'.
 
+B<Singleton pattern:> The WebSocket object is cached in C<< $scope->{'pagi.websocket'} >>.
+If you call C<new()> multiple times with the same scope, you get the same
+WebSocket object back. This ensures consistent state (is_connected, is_closed,
+callbacks) across multiple code paths that may create WebSocket objects from
+the same scope.
+
 =head1 SCOPE ACCESSORS
 
 =head2 scope, path, raw_path, query_string, scheme, http_version
